
from collections import defaultdict
from typing import List
import pandas as pd
from .constants import *




class SexColumnMapper():
    def __init__(self, male_symbol, female_symbol, column_name, other_symbol=None, unknown_symbol=None) -> None:
        self._male_symbol    = male_symbol
        self._female_symbol  = female_symbol
        self._other_symbol   = other_symbol
        self._unknown_symbol = unknown_symbol
        if column_name is None:
            raise ValueError("Must provide non-null column_name argument")
        self._column_name = column_name
    
    def map_cell(self, cell_contents) -> List:
        contents = cell_contents.strip()
        if contents == self._female_symbol:
            return FEMALE_SYMBOL
        elif contents == self._male_symbol:
            return MALE_SYMBOL
        elif contents == self._other_symbol:
            return OTHER_SEX_SYMBOL
        elif contents == self._unknown_symbol:
            return UNKOWN_SEX_SYMBOL
        else:
            print(f"Could not map sex symbol {contents}")
            return UNKOWN_SEX_SYMBOL        

    def preview_column(self, column):
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for index, value in column.items():
            result  = self.map_cell(str(value))
            if result is None:
                dlist.append({"original column contents": value, "sex": "n/a"})
            else:
                dlist.append({"original column contents": value, "sex": result})
        return pd.DataFrame(dlist)
    

    def get_column_name(self):
        return self._column_name


from .hpo_cr import HpoConceptRecognizer
from .hp_term import HpTerm

import re
import pandas as pd
from typing import List
from collections import defaultdict



class HpoExactConceptRecognizer(HpoConceptRecognizer):
    #self._label_to_id_d, self._id_to_primary_label
    def __init__(self, label_to_id, id_to_primary_label):
        super().__init__()
        if not isinstance(label_to_id, dict):
            raise ValueError("label_to_id_d argument must be dictionary")
        if not isinstance(id_to_primary_label, dict):
            raise ValueError("labels_to_primary_label_d argument must be dictionary")
        self._id_to_primary_label = id_to_primary_label
        self._label_to_id = label_to_id
        
        
    def parse_cell(self, cell_contents, custom_d=None)-> List[HpTerm]:
        """_summary_

        Args:
            cell_contents (_type_): description of patient phenotypes. Assumption is that if there are new lines, any given phenotype is on its own line
            custom_d (_type_, optional): User-provided dictionary with mappings to HPO terms. Defaults to None.

        Raises:
            ValueError: must be a string
        """
        if not isinstance(cell_contents, str):
            print(f"Error: cell_contents argument ({cell_contents}) must be string but was {type(cell_contents)} -- coerced to string")
            cell_contents = str(cell_contents)   
            
        lines = self._split_into_lines(cell_contents)
        if custom_d is None:
            # initialize to empty dictionary if this argument is not passed 
            # to avoid needed to check for None in other functions
            custom_d = defaultdict()
        if len(lines) > 1:
            results = []
            for line in lines:
                res = self._parse_line(line=line, custom_d=custom_d)
                results.append(res)
        else:
            # just one line
            return self._parse_line(line=lines[0], custom_d=custom_d)
        
    def _split_into_lines(self, cell_contents):
        """_summary_
        Split a cell into lines and remove white space from beginning and end of each line and transform to lower case
        """
        return cell_contents.split('\n')
    
    def _split_line_into_chunks(self, line):
        """_summary_
        Split a line into chunks and remove white space from beginning and end of each chunk
        """
        delimiters = ',;|/'
        regex_pattern = '|'.join(map(re.escape, delimiters))
        chunks = re.split(regex_pattern, line) 
        return [chunk.strip().lower() for chunk in chunks]
    
    def _parse_chunk(self, chunk, custom_d) -> List[HpTerm]:
        if chunk in self._label_to_id:
            hpo_id = self._label_to_id.get(chunk)
            label = self._id_to_primary_label.get(hpo_id)
            return [HpTerm(id=hpo_id, label=label)]
        else:
            results = []
            # If we get here, we do two things
            # We may have a long text that includes both custom strings and HPO terms/synonyms
            # strategy is to first extract the custom terms and then check the remaining text
            # for HPO terms
            remaining_text = chunk.lower()
            for k, v in custom_d.items():
                key = k.lower()
                # We now try to find a custom item in the potentially longer chunk string
                if key in remaining_text:
                    hpo_label = v
                    hpo_label_lc = hpo_label.lower()    
                    hpo_id = self._label_to_id.get(hpo_label_lc)
                    if hpo_id is None:
                        print(f"Unable to retrieve HPO Id for custom mapping {chunk} -> {hpo_label}")
                        return []
                    results.append(HpTerm(id=hpo_id, label=hpo_label))
                    remaining_text = remaining_text.replace(key, " ")
            # When we get here, we look for HPO terms in the remaining text
            if len(remaining_text) > 5:
                for k, v in self._label_to_id.items():
                    key = k.lower()
                    if key in remaining_text:
                        hpo_id = v
                        hpo_label = self._id_to_primary_label.get(hpo_id)
                        results.append(HpTerm(id=hpo_id, label=hpo_label))
                        remaining_text = remaining_text.replace(key, " ")
                        #print(f"hpo {hpo_label} key {key} remaining {remaining_text}")
            return results
            
    def _parse_line(self, line, custom_d) -> List[HpTerm]:
        """_summary_
        'private' function to parse an entire line or chunk
        The reason we parse lines first is that we are more likely to get complete HPO terms this way
        """
        # remove whitespace, convert to lower case (as )
        content = line.strip().lower() 
        if content is None or len(content) == 0:
            return []
        results = self._parse_chunk(chunk=content, custom_d=custom_d)
        if len(results) > 0:
            return results
        else:
            chunks = self._split_line_into_chunks(content)
            results = []
            for chunk in chunks:
                # Note that chunk has been stripped of whitespace and lower-cased already
                res = self._parse_chunk(chunk=chunk, custom_d=custom_d)
                results.extend(res)
            return results

    def get_term_from_id(self, id) -> HpTerm:
        if not id.startswith("HP:"):
            raise ValueError(f"Malformed HP id '{id}' - must start with HP:")
        if not id in self._id_to_primary_label:
            raise ValueError(f"Could not find id {id} in dictionary")
        label = self._id_to_primary_label.get(id)
        return HpTerm(id=id, label=label)


import unittest
import os
from pyphetools import HpoParser, OptionColumnMapper, HpoConceptRecognizer

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')



class TestOptionMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        cls.severity_d = {'mild': 'Intellectual disability, mild',
             'moderate':'Intellectual disability, moderate',
             'severe': 'Intellectual disability, severe',
             }
        
    def test_hpo_cr_mild(self):
        """We should retrieve Intellectual disability, mild (HP:0001256)"""   
        optionMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("mild")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001256", result.id)
        self.assertEqual("Intellectual disability, mild", result.label) 
        
    def test_hpo_cr_moderate(self):
        """We should retrieve Intellectual disability, moderate (HP:0002342)"""   
        optionMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("moderate")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0002342", result.id)
        self.assertEqual("Intellectual disability, moderate", result.label) 
        
    def test_hpo_cr_severe(self):
        """We should retrieve Intellectual disability, severe (HP:0010864) """   
        optionMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("severe")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0010864", result.id)
        self.assertEqual("Intellectual disability, severe", result.label) 
        
    def test_multiple_items_in_cell_1(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                 'Infantile spasms': 'Infantile spasms',
               'GTC':'Bilateral tonic-clonic seizure',
                'ESES': 'Status epilepticus'}
        seizureMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=seizure_d)
        results = seizureMapper.map_cell("Absence")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0011147", result.id)
        self.assertEqual("Typical absence seizure", result.label) 
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)
    
    def test_multiple_items_in_cell_2(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                 'Infantile spasms': 'Infantile spasms',
               'GTC':'Bilateral tonic-clonic seizure',
                'ESES': 'Status epilepticus'}
        seizureMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=seizure_d)
        results = seizureMapper.map_cell("Absence and GTC")
        self.assertEqual(2, len(results))
        results = sorted(results, key=lambda x: x.id, reverse=True)
        result = results[0]
        self.assertEqual("HP:0011147", result.id)
        self.assertEqual("Typical absence seizure", result.label) 
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)
        result = results[1]
        self.assertEqual("HP:0002069", result.id)
        self.assertEqual("Bilateral tonic-clonic seizure", result.label) 
        self.assertTrue(result.observed)
        self.assertTrue(result.measured)
        
    def test_multiple_items_in_cell_3(self):
        morph_d = {
            'bulbous nasal tip': 'Bulbous nose',
            'prominent lobule of ear':'Large earlobe',
            'tapering fingers':'Tapered finger'
        }
        cell_contents = "Broad forehead, deeply set eyes, ptosis, bulbous nasal tip, micrognathia, prominent lobule of ear, tapering fingers"
        morphMapper = OptionColumnMapper(concept_recognizer=self.hpo_cr, option_d=morph_d)
        results = morphMapper.map_cell(cell_contents)
        self.assertEqual(7, len(results))
        results = sorted(results, key=lambda x: x.id, reverse=True)
        

        

        
        


import unittest
import os
from pyphetools import ThresholdedColumnMapper




class TestThresholdedColumnMapper(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls) -> None:
        cls.mapper = ThresholdedColumnMapper(hpo_id="HP:0032988", 
                                         hpo_label="Persistent head lag", 
                                        threshold=4, 
                                        call_if_above=True)

    def test_threshold_of_3(self):
        res = self.mapper.map_cell(3)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0032988", result.id)
        self.assertEqual("Persistent head lag", result.label)
        self.assertFalse(result.observed)
        
    def test_threshold_of_4(self):
        res = self.mapper.map_cell(4)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0032988", result.id)
        self.assertEqual("Persistent head lag", result.label)
        self.assertFalse(result.observed)
        
    def test_threshold_of_4_point_1(self):
        res = self.mapper.map_cell(4.1)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0032988", result.id)
        self.assertEqual("Persistent head lag", result.label)
        self.assertTrue(result.observed)
        
    def test_threshold_with_na(self):
        """_summary_
        Return cannot be measured if the threshold cannot be read
        """
        res = self.mapper.map_cell("NA")
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0032988", result.id)
        self.assertEqual("Persistent head lag", result.label)
        self.assertFalse(result.measured)


import unittest
import os
from pyphetools import HpoParser, OptionColumnMapper, HpoConceptRecognizer

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')



class TestOptionMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        
        
    def test_mixed(self):
        cell_contents = "Long and thick eyebrows, upper slanted palpebral fissures, anteverted nares, short philtrum"
        other_d = {'upper slanted palpebral fissures': 'Upslanted palpebral fissure'}
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=other_d)
        # sort results in order of HP ids for testing
        # We expect
        # Thick eyebrow (HP:0000574)
        # Upslanted palpebral fissure (HP:0000582)
        # Anteverted nares (HP:0000463)
        # Short philtrum (HP:0000322)
        results = sorted(results, key = lambda x : x.id)
        self.assertEqual(4, len(results))
        self.assertEqual("HP:0000322", results[0].id)
        self.assertEqual("Short philtrum", results[0].label)
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0000463", results[1].id)
        self.assertEqual("Anteverted nares", results[1].label)
        self.assertTrue(results[1].observed)
        self.assertTrue(results[1].measured)
        self.assertEqual("HP:0000574", results[2].id)
        self.assertEqual("Thick eyebrow", results[2].label)
        self.assertTrue(results[2].observed)
        self.assertTrue(results[2].measured)
        self.assertEqual("HP:0000582", results[3].id)
        self.assertEqual("Upslanted palpebral fissure", results[3].label)
        self.assertTrue(results[3].observed)
        self.assertTrue(results[3].measured)
        
    def test_mixed_2(self):
        morph_d = {
            'bulbous nasal tip': 'Bulbous nose',
            'prominent lobule of ear':'Large earlobe',
            'tapering fingers':'Tapered finger'
        }
        cell_contents = "Broad forehead, deeply set eyes, ptosis, bulbous nasal tip, micrognathia, prominent lobule of ear, tapering fingers"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=morph_d)
        # Expect
        # Broad forehead (HP:0000337)
        # Micrognathia (HP:0000347)
        # Bulbous nose (HP:0000414)
        # Deeply set eye (HP:0000490)
        # Ptosis (HP:0000508)
        # Tapered finger (HP:0001182)
        # Large earlobe (HP:0009748)
        self.assertEqual(7, len(results))
        results = sorted(results, key=lambda x: x.id)     
        self.assertEqual("HP:0000337", results[0].id)
        self.assertEqual("Broad forehead", results[0].label) 
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0000347", results[1].id)
        self.assertEqual("Micrognathia", results[1].label) 
        self.assertEqual("HP:0000414", results[2].id)
        self.assertEqual("Bulbous nose", results[2].label) 
        self.assertEqual("HP:0000490", results[3].id)
        self.assertEqual("Deeply set eye", results[3].label) 
        self.assertEqual("HP:0000508", results[4].id)
        self.assertEqual("Ptosis", results[4].label) 
        self.assertEqual("HP:0001182", results[5].id)
        self.assertEqual("Tapered finger", results[5].label) 
        self.assertEqual("HP:0009748", results[6].id)
        self.assertEqual("Large earlobe", results[6].label) 
        
    def test_pair_of_custom_concepts(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                    'Infantile spasms': 'Infantile spasms',
                    'GTC':'Bilateral tonic-clonic seizure',
                    'ESES': 'Status epilepticus'}
        cell_contents = "Absence and GTC"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=seizure_d)
        ## Expect
        # Typical absence seizure (HP:0011147)
        # Bilateral tonic-clonic seizure (HP:0002069)
        self.assertEqual(2, len(results))
        results = sorted(results, key=lambda x: x.id)     
        self.assertEqual("HP:0002069", results[0].id)
        self.assertEqual("Bilateral tonic-clonic seizure", results[0].label) 
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0011147", results[1].id)
        self.assertEqual("Typical absence seizure", results[1].label) 
        self.assertTrue(results[1].observed)
        self.assertTrue(results[1].measured)
            
    def test_get_longest_overlapping(self):
        cell_contents = "Cryptorchidism, micropenis, bilateral talipes equinovarus"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents)
        self.assertEqual(3, len(results))
        

        

        


import unittest
import os
import pandas as pd
from pyphetools import HpoParser, CaseEncoder
HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')
# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"

class TestCaseParse(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        hpparser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        hpo_cr = hpparser.get_hpo_concept_recognizer()
        cls._parser = CaseEncoder(concept_recognizer=hpo_cr, pmid="PMID:1")

    def test_chief_complait(self):
        """
        Expect to get
        HP:0000365 	Hearing impairment 	True 	True
1 	    HP:0001742 	Nasal congestion 	True 	True
2 	    HP:0025267 	Snoring
        """
        vignette = "A 17-mo-old boy presented with progressive nasal obstruction, snoring and hearing loss symptoms when referred to the hospital."
        results = self._parser.add_vignette(vignette=vignette)
        self.assertEqual(3, len(results))
        self.assertTrue(isinstance(results, pd.DataFrame))
        id = results.loc[(results['id'] == 'HP:0000365')]['id'].values[0]
        self.assertEqual("HP:0000365", id)

       



from collections import defaultdict
from typing import List
import pandas as pd
import re


from enum import Enum

AgeEncodingType = Enum('AgeEncodingType', ['YEAR', 'ISO8601', 'YEAR_AND_MONTH', 'CUSTOM'])
ISO8601_REGEX = r"^P(\d+Y)?(\d+M)?(\d+D)?"
# e.g., 14 y 8 m or 8 y
YEAR_AND_MONTH_REGEX = r"(\d+)\s*[Yy]\s*(\d+)\s*[Mm]"
YEAR_REGEX = r"(\d+)\s*[Yy]"
MONTH_REGEX = r"(\d+)\s*[Mm]"


class AgeColumnMapper():
    def __init__(self, ageEncodingType, column_name) -> None:
        self._age_econding = ageEncodingType
        if column_name is None:
            raise ValueError("Must provide non-null column_name argument")
        self._column_name = column_name
    
    
    def map_cell(self, cell_contents) -> List:
        if isinstance(cell_contents, str):
            contents = cell_contents.strip()
        else:
            contents = cell_contents
        if self._age_econding == AgeEncodingType.YEAR:
            try:
                years = int(contents)
                return f"P{years}Y"
            except ValueError as verr:  
                print(f"Could not parse {contents} as integer (year): {verr}")
                return None
        if self._age_econding == AgeEncodingType.YEAR_AND_MONTH:
            try:
                match = re.search(YEAR_AND_MONTH_REGEX, contents)
                if match:
                    years = int(match.group(1))
                    months = int(match.group(2))
                    return f"P{years}Y{months}M"
                match = re.search(YEAR_REGEX, contents)
                if match:
                    years = int(match.group(1))
                    return f"P{years}Y"
                match = re.search(MONTH_REGEX, contents)
                if match:
                    months = int(match.group(1))
                    return f"P{months}M"
            except ValueError as verr:  
                print(f"Could not parse {contents} as year/month: {verr}")
                return None      
    
        elif self._age_econding == AgeEncodingType.ISO8601:
            match = re.search(ISO8601_REGEX, contents)
            if match:
                return contents
            else:
                print(f"Could not parse {contents} as ISO8601 period")
                return None
        elif self._age_econding == AgeEncodingType.CUSTOM:
            raise ValueError("TODO NOT IMPLEMENTED YET")   



    def preview_column(self, column):
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        preview_d = defaultdict(list)
        for index, value in column.items():
            preview_d[value]  = self.map_cell(str(value))
        dlist = []
        for k, v in preview_d.items():
            if v is None:
                dlist.append({"original column contents": k, "age": "n/a"})
            else:
                dlist.append({"original column contents": k, "age": v})
        return pd.DataFrame(dlist)
    
    def get_column_name(self):
        return self._column_name
    
    @staticmethod
    def by_year(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.YEAR, column_name=column_name)
    
    
    @staticmethod
    def by_year_and_month(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.YEAR_AND_MONTH, column_name=column_name)
    
    @staticmethod
    def iso8601(column_name):
        return AgeColumnMapper(ageEncodingType=AgeEncodingType.ISO8601, column_name=column_name)

   


import pandas as pd


class HpTerm:
    def __init__(self, id, label, observed=True, measured=True):
        if id is None or len(id) == 0 or not id.startswith("HP"):
            raise ValueError(f"invalid id argument: '{id}'")
        if label is None or len(label) == 0:
            raise ValueError(f"invalid label argument: '{label}'")
        self._id = id
        self._label = label
        self._observed = observed
        self._measured = measured
        
    def __hash__(self):
        return hash((self._id, self._label, self._observed, self._measured))
        
    @property
    def id(self):
        return self._id
    
    @property
    def label(self):
        return self._label
    
    @property
    def observed(self):
        return self._observed
    
    @property
    def measured(self):
        return self._measured
    
    @property
    def display_value(self):
        if not self._measured:
            return "not measured"
        if not self._observed:
            return "excluded"
        else:
            return "observed"
    
    @property
    def hpo_term_and_id(self):
        return f"{self._label} ({self._id})"
    
    def __str__(self) -> str:
        if not self._measured:
            return f"not measured: {self._label} ({self._id})"
        elif not self._observed:
            return f"excluded: {self._label} ({self._id})"
        else:
            return f"{self._label} ({self._id})"
    
    def to_string(self):
        return self.__str__()


    @staticmethod
    def term_list_to_dataframe(hpo_list):
        if not isinstance(hpo_list, list):
            raise ValueError(f"hpo_list argument must be a list but was {type(hpo_list)}")
        if len(hpo_list) > 0:
            hpo1 = hpo_list[0]
            if not isinstance(hpo1, HpTerm):
                raise ValueError(f"hpo_list argument must consist of HpTerm objects but had {type(hpo1)}")
        if len(hpo_list) == 0:
            return pd.DataFrame(columns=['Col1', 'Col2', 'Col3'])
        items = []
        for hp in hpo_list:
            d = { "id": hp._id, "label": hp._label, "observed":hp._observed, "measured": hp._measured }
            items.append(d)
        return pd.DataFrame(items)


import phenopackets 
from .variant import Variant
from .constants import *


class Individual:
    """
    A class to represent one individual of the cohort
    """
    def __init__(self, individual_id, sex, age, hpo_terms, variant_list=None, disease_id=None, disease_label=None):
        """
        Represents all of the data we will transform into a single phenopacket
        """
        if not isinstance(individual_id, str):
            raise ValueError(f"individual_id argument must be string but was {type(individual_id)}")
        self._individual_id = individual_id
        self._sex = sex
        self._age = age
        self._hpo_terms = hpo_terms
        self._variant_list = variant_list
        self._disease_id = disease_id
        self._disease_label = disease_label
        
    @property
    def id(self):
        return self._individual_id
    
    @property
    def sex(self):
        return self._sex
    
    @property
    def age(self):
        return self._age
    
    @property
    def hpo_terms(self):
        return self._hpo_terms
    
    @property
    def variant_list(self):
        return self._variant_list
    
    def to_ga4gh_phenopacket(self, metadata=None, phenopacket_id=None):
        """_summary_
        Transform the data into GA4GH Phenopacket format
        Returns:
            _type_: _description_
        """
        php = phenopackets.Phenopacket()
        if phenopacket_id is None:
            php.id = self._individual_id
        else:
            php.id = phenopacket_id
        php.subject.id = self._individual_id
        if self._sex == MALE_SYMBOL:
            php.subject.sex = phenopackets.Sex.MALE
        elif self._sex == FEMALE_SYMBOL:
            php.subject.sex = phenopackets.Sex.FEMALE
        elif self._sex == OTHER_SEX_SYMBOL:
            php.subject.sex = phenopackets.Sex.OTHER
        elif self._sex == UNKOWN_SEX_SYMBOL:
            php.subject.sex = phenopackets.Sex.UNKNOWN
        if self._age is not None:
            php.subject.time_at_last_encounter.age.iso8601duration = self._age
        if isinstance(self._hpo_terms, list):
            for hp in self._hpo_terms:
                if not hp.measured:
                    continue
                pf = phenopackets.PhenotypicFeature()
                pf.type.id = hp.id
                pf.type.label = hp.label
                if not hp.observed:
                    pf.excluded = True
                if self._age is not None:
                    pf.onset.age.iso8601duration = self._age
                php.phenotypic_features.append(pf)
        elif isinstance(self._hpo_terms, dict):
            for age_key, hpoterm_list in self._hpo_terms.items():
                for hp in hpoterm_list:
                    if not hp.measured:
                        continue
                    pf = phenopackets.PhenotypicFeature()
                    pf.type.id = hp.id
                    pf.type.label = hp.label
                    if not hp.observed:
                        pf.excluded = True
                    if age_key.startswith("P"):
                        # Note sometimes we have no age, then use N/A -- TODO think of robust way to do this
                        pf.onset.age.iso8601duration = age_key
                    php.phenotypic_features.append(pf)
        if len(self._variant_list) > 0:
            interpretation = phenopackets.Interpretation()
            interpretation.id = self._individual_id
            interpretation.progress_status = phenopackets.Interpretation.ProgressStatus.SOLVED
            if self._disease_id is not None and self._disease_label is not None:
                interpretation.diagnosis.disease.id = self._disease_id
                interpretation.diagnosis.disease.label = self._disease_label
            for var in self._variant_list:
                genomic_interpretation = phenopackets.GenomicInterpretation()
                genomic_interpretation.subject_or_biosample_id = self._individual_id
                # by assumption, variants passed to this package are all causative
                genomic_interpretation.interpretation_status = phenopackets.GenomicInterpretation.InterpretationStatus.CAUSATIVE
                genomic_interpretation.variant_interpretation.CopyFrom(var.to_ga4gh())
                interpretation.diagnosis.genomic_interpretations.append(genomic_interpretation)
            php.interpretations.append(interpretation)
        if metadata is not None:
            php.meta_data.CopyFrom(metadata)
        return php


import unittest
import os
from pyphetools import HpoParser, CustomColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

class TestCustomColumnMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        
        
    def test_hpo_cr_non_null(self):
        """sanity check"""
        self.assertIsNotNone(self.hpo_cr)
        
    def test_hpo_cr_ataxia(self):
        """We should retrieve Ataxia HP:0001251"""   
        neuroMapper = CustomColumnMapper(concept_recognizer=self.hpo_cr)
        results = neuroMapper.map_cell("ataxia")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001251", result.id)
        self.assertEqual("Ataxia", result.label)
        
    def test_hpo_cr_spastic_paraplegia(self):
        """We should retrieve Spastic paraplegia (HP:0001258)"""   
        neuroMapper = CustomColumnMapper(concept_recognizer=self.hpo_cr)
        results = neuroMapper.map_cell("spastic paraplegia")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001258", result.id)
        self.assertEqual("Spastic paraplegia", result.label)   
        
        
    def test_hpo_cr_multiple_concepts_with_custom_map(self):
        text = 'spasticity; nerve conduction and EMG studies with abnormal findings "remarkable for the failure to activate the leg muscles due to an upper motor neuron pattern of aberrant motor unit potential firing rates. These findings are consistent with dysfunction of the corticospinal pathways rather than a lower motor unit."'
        neuroMapper = CustomColumnMapper(concept_recognizer=self.hpo_cr)
        results = neuroMapper.map_cell(text)
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001257", result.id)
        self.assertEqual("Spasticity", result.label)   
    
    def test_hpo_cr_multiple_concepts_no_custom_map(self):
        text = 'spasticity; nerve conduction and EMG studies with abnormal findings "remarkable for the failure to activate the leg muscles due to an upper motor neuron pattern of aberrant motor unit potential firing rates. These findings are consistent with dysfunction of the corticospinal pathways rather than a lower motor unit." Significant low extremity weakness.'
        neuro_exam_custom_map = {'low extremity weakness': 'Lower limb muscle weakness',  
                         'unstable gait': 'Unsteady gait',
                         'dysfunction of the corticospinal pathways':'Upper motor neuron dysfunction',
                        }
        neuroMapper = CustomColumnMapper(concept_recognizer=self.hpo_cr, custom_map_d=neuro_exam_custom_map)
        results = neuroMapper.map_cell(text)
        self.assertEqual(3, len(results))
        term_d = dict([(hpo_term.id, hpo_term.label) for hpo_term in results])
        self.assertTrue("HP:0001257" in term_d)
        self.assertEqual("Spasticity", term_d.get("HP:0001257"))
        self.assertTrue("HP:0007340" in term_d)
        self.assertEqual("Lower limb muscle weakness", term_d.get("HP:0007340"))
        self.assertTrue("HP:0002493" in term_d)
        self.assertEqual("Upper motor neuron dysfunction", term_d.get("HP:0002493"))
        